#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <utils/PugsAssert.hpp>
#include <utils/SubArray.hpp>
#include <utils/Types.hpp>

// Instantiate to ensure full coverage is performed
template class SubArray<int>;

// clazy:excludeall=non-pod-global-static

TEST_CASE("SubArray", "[utils]")
{
  Array<int> a(10);
  REQUIRE(a.size() == 10);

  SECTION("shared values")
  {
    SubArray sub_a{a, 0, 10};
    for (size_t i = 0; i < sub_a.size(); ++i) {
      sub_a[i] = 2 * i;
    }

    REQUIRE(((a[0] == 0) and (a[1] == 2) and (a[2] == 4) and (a[3] == 6) and (a[4] == 8) and (a[5] == 10) and
             (a[6] == 12) and (a[7] == 14) and (a[8] == 16) and (a[9] == 18)));

    for (size_t i = 0; i < a.size(); ++i) {
      a[i] = (i + 1) * (2 * i + 1);
    }

    REQUIRE(((sub_a[0] == 1) and (sub_a[1] == 6) and (sub_a[2] == 15) and (sub_a[3] == 28) and (sub_a[4] == 45) and
             (sub_a[5] == 66) and (sub_a[6] == 91) and (sub_a[7] == 120) and (sub_a[8] == 153) and (sub_a[9] == 190)));
  }

  SECTION("sub array")
  {
    a.fill(0);
    SubArray sub_a{a, 5, 5};
    for (size_t i = 0; i < sub_a.size(); ++i) {
      sub_a[i] = i + 1;
    }

    REQUIRE(((a[0] == 0) and (a[1] == 0) and (a[2] == 0) and (a[3] == 0) and (a[4] == 0) and (a[5] == 1) and
             (a[6] == 2) and (a[7] == 3) and (a[8] == 4) and (a[9] == 5)));

    for (size_t i = 0; i < a.size(); ++i) {
      a[i] = (i + 1) * (2 * i + 1);
    }

    REQUIRE(((sub_a[0] == 66) and (sub_a[1] == 91) and (sub_a[2] == 120) and (sub_a[3] == 153) and (sub_a[4] == 190)));
  }

#ifndef NDEBUG
  SECTION("errors")
  {
    a.fill(0);
    SubArray<int> sub_a{a, 5, 5};
    for (size_t i = 0; i < sub_a.size(); ++i) {
      sub_a[i] = i + 1;
    }

    REQUIRE_THROWS_AS(sub_a[5], AssertError);
  }
#endif   // NDEBUG
}
